(in-package medic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User parameters
;;;

(defvar *exclusion-strategy* 'parallel)
(defvar *exclusion-strategies* '(sequential-pairwise parallel))
(deftype exclusion-strategy ()
  `(member ,@*exclusion-strategies*))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Exclusion clauses
;;;

(defun exclusion-clauses ()
  (if *show-stages* (format t "~&Entering exclusion-clauses~%"))
  (cond ((eq *action-representation* 'bitwise)
	 ;; The *split-table* was already computed in act-rep-clauses
	 (values '() '()))
	((eq *exclusion-strategy* 'sequential-pairwise)
	 (sequential-pairwise-excl-clauses))
	((eq *exclusion-strategy* 'parallel)
	 (parallel-excl-clauses))
	(t
	 (error "Bad value for *exclusion-strategy* ~s" *exclusion-strategy*))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sequential pairwise exclusion
;;;

;; uses the global variable *ground-actions* to get a list of actions
;; grounded for operators.
;; Returns a TNPF and substitutions to handle sequential-pairwise exclusion.
;; (defun sequential-pairwise-excl-clauses ()
;;   (values
;;    (exactly-one-sequential-pairwise *ground-actions*)
;;    nil))

(defun sequential-pairwise-excl-clauses ()
  (assert (not (eq *action-representation* 'bitwise)))
  (values
   (cond ((member *action-representation* '(regular simple-split overloaded-split))
	  (eliminate-instantiated-non-removable-types-clauses-sprop
	   (exactly-one-sequential-pairwise *ground-actions*)))
	 ;; For simple split overloading, we have 3 possibilities:
	 ;; * do the usual thing: 'overloaded-split
	 ;; * require good assignment for all variables: 'overloaded-split-factored
	 ;; * require good assignment only for some variables: 'overloaded-split-partially-factored
	 ((member *action-representation* '(simple-split-factored
					    overloaded-split-factored))
	  (eliminate-instantiated-non-removable-types-clauses-sprop
	   (sequential-pairwise-excl-clauses-factored)))
	 (t
	  (error "bad value for *action-representation* ~s" *action-representation*)))
   nil))

(defun exactly-one-sequential-pairwise (possibilities)
  (let ((no-two-simultaneously
	 ;; no two possibilities are true
	 (mapcon #'(lambda (l1)
		     (mapcar #'(lambda (e2)
				 (list :or (sprop-negate (car l1)) (sprop-negate e2)))
			     (cdr l1)))
		 possibilities)))
    (case *frame-strategy*
      (explanatory no-two-simultaneously)
      (classical (if (factored-splitting-p)
		     '()
		   (list (cons :or possibilities))))
      (otherwise
       (error "Bad value for *frame-strategy* ~s" *frame-strategy*)))))


;; works for both operator and overloaded splitting
(defun sequential-pairwise-excl-clauses-factored ()
  (let* ((maxarity (max-arity *operators*))
	 (actionat-props (append
			   ;; list of "action-at" sprops
			   (mapcar #'(lambda (split-entry)
				       (first (cdr split-entry)))
				   *split-table*)
			   ;; list of unsplit operators (= sprops)
			   *split-untouched*))
	 (op-one (exactly-one-sequential-pairwise
		  (remove-duplicates actionat-props :test #'sprop-equal))))
    ;; (format t "actionat-props ~s~%op-one ~s~%" actionat-props op-one)
    (cond ((eq *action-representation* 'overloaded-split-factored)
	   (let
	       ;; one-at-a-time clauses
	       ((arg-ones
		 (loop for i from 1 to maxarity
		     collect
		       (exactly-one-sequential-pairwise
			(mapcar #'(lambda (obj)
				    (overloaded-split-argnum-sprop i obj :during))
				;; Could instantiate just for the objects
				;; that can be the ith argument, rather than
				;; for every object in the world.  Don't
				;; forget unsplit unary operators!
				*objects*)))))
	     (apply #'nconc op-one arg-ones)))
;; 	  ((eq *action-representation* 'overloaded-split-partially-factored)
;; 	   (error "not tested yet")
;; 	   (let* ((arg-ones (error "copy from above? restore to outer let* ?"))
;; 		  (ops-by-argno
;; 		   (let ((result '()))
;; 		     (dolist (op *operators*)
;; 		       (let* ((act (op-action op))
;; 			      (arity (length (cdr act)))
;; 			      (arity-assoc (assoc arity result)))
;; 			 (if arity-assoc
;; 			     (push (car act) (cdr arity-assoc))
;; 			   (push (list arity (car act)) result))))
;; 		     result))
;; 		  (arg-ones-by-arity
;; 		   (mapcar #'(lambda (num-and-acts)
;; 			       (let ((num (car num-and-acts))
;; 				     (acts (cdr num-and-acts)))
;; 				 (make-implies
;; 				  (cons :or
;; 					(mapcar #'(lambda (act)
;; 						    (make-actionat-sprop act nil))
;; 						acts))
;; 				  (cons :and
;; 					(butlast arg-ones (- maxarity num))))))
;; 			   ops-by-argno)))
;; 	     (nconc op-one arg-ones-by-arity)))
	  ((eq *action-representation* 'simple-split-factored)
	   ;; op-one is OK.
	   (let ((arg-ones
		  (mapcan
		   #'(lambda (op)
		       (let ((obj-lst (op-obj-lst op)))
			 (loop for i from 2 to (length obj-lst)
			     collect
			       (exactly-one-sequential-pairwise
				(mapcar #'(lambda (obj)
					    (simple-split-argnum-sprop
					     op i obj :during))
					;; Wrong to use *objects*, as that
					;; introduces propositions prohibited
					;; by obj-lst.
					(nth (1- i) obj-lst))))))
		   *operators*)))
	     (apply #'nconc op-one arg-ones)))
	  (t
	   (error "*action-representation* must be 'overloaded-split-factored not ~s" *action-representation*)))))


;; Old implementation
;; (defun sequential-pairwise-excl-clauses-factored ()
;;   (let* ((maxarity (reduce #'max (mapcar #'op-arity *operators*))))
;; 	 ;; one-at-a-time clauses
;; 	 (arg-ones (loop for i from 1 to maxarity
;; 		       collect
;; 			 (exactly-one-sequential-pairwise
;; 			  (mapcar #'(lambda (obj)
;; 				      (overloaded-split-argnum-sprop i obj nil))
;; 				  *objects*))))
;; 	 (ground-actions (nconc (mapcar #'second *split-table*)
;; 				*split-untouched*))
;; 	 (op-one (exactly-one-sequential-pairwise ground-actions)))
;;     (cond ((eq *action-representation* 'overloaded-split-factored)
;; 	   (apply #'nconc op-one arg-ones))
;; 	  ((eq *action-representation* 'overloaded-split-partially-factored)
;; 	   (let* ((ops-by-argno
;; 		   (let ((result '()))
;; 		     (dolist (op *operators*)
;; 		       (let* ((act (op-action op))
;; 			      (arity (length (cdr act)))
;; 			      (arity-assoc (assoc arity result)))
;; 			 (if arity-assoc
;; 			     (push (car act) (cdr arity-assoc))
;; 			   (push (list arity (car act)) result))))
;; 		     result))
;; 		  (arg-ones-by-arity
;; 		   (mapcar #'(lambda (num-and-acts)
;; 			       (let ((num (car num-and-acts))
;; 				     (acts (cdr num-and-acts)))
;; 				 (make-implies
;; 				  (cons :or
;; 					(mapcar #'(lambda (act)
;; 						    (make-actionat-sprop act nil))
;; 						acts))
;; 				  (cons :and
;; 					(butlast arg-ones (- maxarity num))))))
;; 			   ops-by-argno)))
;; 	     (nconc op-one arg-ones-by-arity)))
;; 	  (t
;; 	   (error "*action-representation* must be 'overloaded-split-{fully,partially}-factored not ~s" *action-representation*)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Parallel Exclusion
;;;

(defun parallel-excl-clauses ()
  (values
   (eliminate-instantiated-non-removable-types-clauses-sprop
    (exactly-one-parallel *ground-action-precond-lits*
			  *ground-action-effect-lits*))
   nil))


;; Returns t if the two different actions whose grounded precondition and
;; effect literals are given are mutually exclusive.
(defun parallel-exclusive-p (name-precond-lits name-effect-lits)
  (assert (every #'sprop-p (cdr name-precond-lits)))
  (assert (every #'sprop-p (cdr name-effect-lits)))
  (not (null (intersection-sprop-equal (cdr name-precond-lits)
				       (cdr name-effect-lits)))))

;; "exactly-one-parallel" is a horrible name, since this doesn't enforce that
;;   constraint at all.
;; name-pre-lits is a list of pairs of (ground-action . literals-in-precondition)
;; name-eff-lits is a list of pairs of (ground-action . literals-in-effects)
;; The two lists must be in the same order!
;; (It would be simpler to take just a single list...)
(defun exactly-one-parallel (name-pre-lits name-eff-lits)
  (assert (eq 'explanatory *frame-strategy*))
  (let ((no-conflicts
	 ;; no two conflicting possibilities are true
	 (mapcon
	  #'(lambda (pre-list eff-list)
	      (let ((name-pre1 (car pre-list))
		    (name-eff1 (car eff-list)))
		;; make sure name-pre-lits and name-eff-lits in same order
		(assert (sprop-p (car name-pre1)))
		(assert (sprop-p (car name-eff1)))
		(assert (sprop-equal (car name-pre1) (car name-eff1)))
		(mapcan #'(lambda (name-pre2 name-eff2)
			    ;; Shouldn't have two identical names
			    (assert (not (sprop-equal (car name-pre1) (car name-pre2))))
			    (if (or (parallel-exclusive-p name-pre1 name-eff2)
				    (parallel-exclusive-p name-pre2 name-eff1))
				(list (list :or (sprop-negate (car name-pre1))
					    (sprop-negate (car name-pre2))))
			      '()))
			(cdr pre-list) (cdr eff-list))))
	  name-pre-lits name-eff-lits)))
    ;; No need for this, because we are certainly in the effectual case.
    ;; ;; at least one possibility is true
    ;; (cons :or (mapcar #'car name-pre-lits))
    no-conflicts))
