(in-package medic)

;; In addition to
;;   (multi-regress)
;;   (multi-flags-regress)
;; consider modifying some or all of the following variables for better
;; code coverage (see section "User parameters" of each file for details).
;;   *find-minimal-plan*, *infer-types*
;;   *partial-action-method*, *exists-strategy*
;;   *cnf-solver*, *external-simplifier*


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Regression tests
;;;

;; If an assertion fails, just use "c" to continue past it.
(defun regress ()
  ;; These assertions shouldn't get commented out ever.  How to so indicate?
  (assert (tiny))
  (assert (tiny2))
  (assert (tiny-when))
  (assert (tiny-eliminate))
  (assert (tiny-exists))
  (assert (or (not (eq 'parallel *exclusion-strategies*)) (tiny2p)))
  (assert (tiny-nosplit))
  (assert (tiny-split))
  (assert (suss-1-2op-2obj))
  (assert (suss))
  (assert (ds1))
  (assert (ds3))
  (assert (ds5))
  (assert (ferry3))
  (assert (ferry))
  (assert (solve-find-problem 'get-paid-strips))
  (assert (solve-find-problem 'r-test1))
  (assert (solve-find-problem 'uget-paid))
  (assert (att-log0))
  (assert (att-log1))
  (assert (att-log2))
  (assert (wheel15))
  (assert (wheel12))
  (assert (wheel11))
  (assert (fridge-fixa))
  (assert (fridge-fixb))
  (assert (tire1))
  (assert (tire3))
  (assert (solve-find-problem 'mcd-tiny))
  ;; Takes way too long
  ;; (assert (solve-find-problem 'mcd-sussman-anomaly))
  ;; add more here
  (format t "Regression test succeeded for flags ~a~%" (flags-brief))
  )

;; Don't use this; use multi-flags-regress instead.
(defun multi-regress (&optional (frames *frame-strategies*)
				(act-reps *action-representations*)
				(exclusions *exclusion-strategies*)
				(types *types-strategies*))
  (assert (notany #'null (list exclusions frames act-reps types)))
  (assert (every #'listp (list exclusions frames act-reps types)))
  (dolist (*frame-strategy* frames)
    (dolist (*action-representation* act-reps)
      (dolist (*exclusion-strategy* exclusions)
	(dolist (*types-strategy* types)
	  (if (check-flags 'no-err)
	      (regress)))))))



;; Perhaps use multi-flags-solve-problem here.

(defvar *regression-problems*
  '(tiny
    tiny2
    tiny-when
    tiny-eliminate
    tiny-exists
    ;; note tiny2p does not appear here
    tiny-nosplit
    tiny-split
    prodigy-sussman-2op-1-2obj
    prodigy-sussman-3
    d1s1-1
    d1s1-3
    d1s1-5
    ferry3
    ferry7
    get-paid-strips
    r-test1
    uget-paid
    att-logistics0
    att-logistics1
    att-logistics2
    wheel-domain15
    wheel-domain12
    wheel-domain11
    fixa-strips
    fixb-strips
    fix1-strips
    fix3-strips
    mcd-tiny
    ;; Takes way too long
    ;; mcd-sussman-anomaly
    ;; add more here
    ))


;; Like regress, but fewer hints are given about correct length (except max).
;; To make this not skip any files, do (medic-clean-directory) first.

;; There's a problem with max-steps; I can't really set it sensibly for
;; an entire problem, as it varies depending on encoding.  For now, just
;; let it be automatically set, so we don't test the binary search.

;; multi-flags-solve-problem checks that a plan is returned unless
;; *dont-solve* is set; set that variable if you just want to generate CNF.
(defun multi-flags-regress (&optional
				      (problemnames *regression-problems*)
				      max-steps
				      (flag-lst *flags-list-maximal-types*))
  (dolist (problemname problemnames)
    (multi-flags-solve-problem problemname nil max-steps 4 2 flag-lst)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Testing
;;;

;;; The canoncial cases:  Sussman, Tire, D1S1

(defun suss ()
  (solve-find-problem 'prodigy-sussman-3
		      6 (if *find-minimal-plan* 5 6)
		      "sussman"))
(defun suss-1 ()
  (solve-find-problem 'Prodigy-sussman-1
		      2 (if *find-minimal-plan* 1 2)
		      "sussman"))
(defun suss-1-2op ()
  (solve-find-problem 'prodigy-sussman-2op-1 1 2 "sussman-2op"))
(defun suss-1-2obj ()
  (solve-find-problem 'prodigy-sussman-1-2obj
		      2 (if *find-minimal-plan* 1 2)
		      "sussman"))
(defun suss-1-2op-2obj ()
  (solve-find-problem 'prodigy-sussman-2op-1-2obj
		      2 (if *find-minimal-plan* 1 2)
		      "sussman-2op"))
(defun suss-2 ()
  (solve-find-problem 'prodigy-sussman-2
		      4 (if *find-minimal-plan* 3 4)
		      "sussman"))
(defun suss-2.5 ()
  (solve-find-problem 'prodigy-sussman-2.5
		      5 (if *find-minimal-plan* 4 5)
		      "sussman"))
(defun suss-3 ()
  (solve-find-problem 'Prodigy-sussman-3
		      6 (if *find-minimal-plan* 5 6)
		      "sussman"))
(defun suss-3-small ()
  (solve-find-problem 'prodigy-sussman-smaller
		      6 (if *find-minimal-plan* 5 6)
		      "sussman-smaller"))
(defun suss-3-noobj ()
  (solve-find-problem 'prodigy-sussman-noobj
		      6 (if *find-minimal-plan* 5 6)
		      "sussman-noobj"))
(defun suss-3-neq ()
  (solve-find-problem 'prodigy-sussman-neq 6
		      (if *find-minimal-plan* 5 6)
		      "sussman-neq"))

(defun suss-when ()
  (solve-find-problem 'SUSSMAN-ANOMALY
		      3 (if *find-minimal-plan* 2 3)
		      "blocks-world"))

;; Don't know how many steps for these
(defun tire ()
  (solve-find-problem 'FIXIT-strips 25 20 "tire"))

(defun ds1 ()
  (solve-find-problem 'D1S1-1 1 1 "D1S1"))
(defun ds3 ()
  (solve-find-problem 'D1S1-3 3 3 "D1S1"))
(defun ds5 ()
  (solve-find-problem 'D1S1-5 5 5 "D1S1"))

(defun ds1a ()
  (solve-find-problem 'd1s1-prob-a 1 1 "d1s1a"))
(defun ds1b ()
  (solve-find-problem 'd1s1-prob-b 1 1 "d1s1b"))
(defun ds3a ()
  (solve-find-problem 'd1s3-prob-a 3 3 "d1s3a"))


;; ATT-LOGISTICS WORLD

(defun att-log0 ()
  (solve-find-problem 'att-logistics0 3 3 "logistics"))
(defun att-log1 ()
  (solve-find-problem 'att-logistics1 6 6 "logistics"))
(defun att-log2 ()
  (solve-find-problem 'att-logistics2 9 9 "logistics"))
(defun att-log3 ()
  (solve-find-problem 'att-logistics3 9 9 "logistics"))
(defun att-logA ()
  (solve-find-problem 'att-logistics-A 10 10 "logistics"))

(defun att-log0a ()			; like att-log0, but remove trucks, POs
  (solve-find-problem 'att-log0a 3 3 "att-log0a"))
(defun att-log0b ()			; like att-log0a, but one step only
  (solve-find-problem 'att-log0b 1 1 "att-log0b"))
(defun att-log0c ()			; like att-log0a, but split AT in two
  (solve-find-problem 'att-log0c 3 3 "att-log0c"))

(defun att-log1a ()			; eliminates bos-po,truck from att-log1
  (solve-find-problem 'att-log1a 6 6 "att-log1a"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Other tests
;;;

(defun tiny ()
  (solve-find-problem 'tiny 1 1 "tiny"))
(defun tiny2 ()
  (solve-find-problem 'tiny2 2 2 "tiny2"))
(defun tiny6 ()
  (solve-find-problem 'tiny6 6 6 "tiny6"))
(defun tiny-when ()
  (solve-find-problem 'tiny-when 1 1 "tiny-when"))
(defun tiny-eliminate ()
  (solve-find-problem 'tiny-eliminate 1 1 "tiny-eliminate"))
(defun tiny-exists ()
  (solve-find-problem 'tiny-exists 1 1 "tiny-exists"))
;; This can't actually be split!
(defun tiny-nosplit ()
  (solve-find-problem 'tiny-nosplit 2 2 "tiny-nosplit"))

;; parallel tests
(defun tiny2p ()			; parallel
  (solve-find-problem 'tiny2 1 1 "tiny2"))


(defun ds (n)
  (solve-find-problem (intern (format nil "D1S1-~d" n))
			   n (if *find-minimal-plan* (1- n) n) "D1S1"))
(defun ds4 ()
  (solve-find-problem 'd1s4-prob 4 4 "d1s4"))
;; ds4a has one type removed.
(defun ds4a ()
  (solve-find-problem 'd1s4-prob-a 4 4 "d1s4"))





(defun ferry ()
  (solve-find-problem 'ferry7 7 7 "ferry-domain"))
(defun ferry7 ()
  (solve-find-problem 'ferry7 7 7 "ferry-domain"))
(defun ferry3 ()
  (solve-find-problem 'ferry3 3 3 "ferry-domain"))

(defun wheel1 ()
  (solve-find-problem 'fix-strips1 9 9 "tire-world"))
(defun wheel2 ()
  (solve-find-problem 'fix-strips2 14 14 "tire-world"))
(defun wheel3 ()
  (solve-find-problem 'fix-strips3 22 22 "tire-world"))
(defun wheel4 ()
  (solve-find-problem 'fix-strips4 22 22 "tire-world"))

(defun wheel11 ()
  (solve-find-problem 'wheel-domain11 7 7 "wheel11"))
(defun wheel12 ()
  (solve-find-problem 'wheel-domain12 5 5 "wheel12"))
(defun wheel13 ()
  (solve-find-problem 'wheel-domain13 3 3 "wheel13"))
(defun wheel13a ()
  (solve-find-problem 'wheel-domain13-simpler 3 3 "wheel13"))
(defun wheel14 ()
  (solve-find-problem 'wheel-domain14 2 2 "wheel14"))
(defun wheel15 ()
  (solve-find-problem 'wheel-domain15 2 2 "wheel14"))
(defun wheel16 ()
  (solve-find-problem 'wheel-domain16 1 1 "wheel16"))

(defun ms-wheel13 ()
  (multi-flags-solve-problem 'wheel-domain13 "wheel13" 3 3 3150))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Splitting tests
;;;

;; suss-1-2op-2obj is a good splitting test.

;; Ours
(defun tiny-split ()
  (solve-find-problem 'tiny-split 2 2 "tiny-split"))
(defun split1 ()
  (solve-find-problem 'split1-prob 1 1 "split-1"))
(defun split2 ()
  (solve-find-problem 'split2-prob 2 2 "split-2"))
(defun tiny-hanoi ()			; presently identical to hanoi1
  (solve-find-problem 'hanoi2-2 1 1 "hanoi2"))


;; fixa doesn't require last 6 actions of fixb.
(defun fridge-fixa ()
  (solve-find-problem 'fixa-strips 7 7 "fridge"))
(defun fridge-fixb ()
  (solve-find-problem 'fixb-strips 13 13 "fridge"))

;; omits operators from FIXA
(defun fridgeaa ()
  (solve-find-problem 'fixa-1 7 7 "fridgea-alt1"))
;; contains start-fridge which isn't in alt
(defun fridgeaa2 ()
  (solve-find-problem 'fixa-2 7 7 "fridgea-alt2"))
;; contains screw which isn't in alt2; identical to FIXA, for paranoia's sake.
(defun fridgeaa3 ()
  (solve-find-problem 'fixa-3 7 7 "fridgea-alt3"))
;; just needs a screw operation; for testing action and fluent with same name
(defun fridgeaa4 ()
  (solve-find-problem 'fixa-4 1 1 "fridgea-alt4"))
;; like alt4, but only has screw, unscrew operations
(defun fridgeaa5 ()
  (solve-find-problem 'fixa-5 1 1 "fridgea-alt5"))
;; like alt5, but only has screw operations, omits some other fluents/constants
;; This one works!
(defun fridgeaa6 ()
  (solve-find-problem 'fixa-6 1 1 "fridgea-alt6"))
;; like alt5, but omtis some fluents/constants
(defun fridgeaa7 ()
  (solve-find-problem 'fixa-7 1 1 "fridgea-alt5"))
;; like alt7, but omit the other 3 screws.
;; This one works!
(defun fridgeaa8 ()
  (solve-find-problem 'fixa-8 1 1 "fridgea-alt5"))
;; like alt7, but omit 2 other screws
(defun fridgeaa9 ()
  (solve-find-problem 'fixa-9 1 1 "fridgea-alt5"))


(defun tire5 ()
  (solve-find-problem 'FIX5-strips 1 1 "tire"))
(defun tire4 ()
  (solve-find-problem 'FIX4-strips 4 4 "tire"))
;; can't invoke inflate in this problem.
(defun tire3 ()
  (solve-find-problem 'FIX3-strips 6 6 "tire"))
(defun tire2 ()
  (solve-find-problem 'FIX2-strips 3 3 "tire"))
(defun tire1 ()
  (solve-find-problem 'FIX1-strips 5 5 "tire"))

(defun monkey3 ()
  (solve-find-problem 'monkey-test3 30 1 "monkey"))
(defun monkey2 ()
  (solve-find-problem 'monkey-test2 30 1 "monkey"))
(defun monkey1 ()
  (solve-find-problem 'monkey-test1 30 1 "monkey"))

(defun bwdti3 ()
  (solve-find-problem 'TOWER-INVERT3 8 1 "blocks-world"))
(defun bwdti4 ()
  (solve-find-problem 'TOWER-INVERT4 8 1 "blocks-world"))
(defun bwbw1 ()
  (solve-find-problem 'simple-block1 3 3 "blocks-world"))
(defun bwbw2 ()
  (solve-find-problem 'simple-block2 4 4 "blocks-world"))
(defun bwbw3 ()
  (solve-find-problem 'simple-block3 8 1 "blocks-world"))
(defun hanoi1 ()
  (solve-find-problem 'hanoi-strips1 1 1 "hanoi"))
(defun hanoi2 ()
  (solve-find-problem 'hanoi-strips2 8 1 "hanoi")) ; thistle
(defun hanoi3 ()
  (solve-find-problem 'hanoi-strips3 8 1 "hanoi")) ; arnica
(defun ms2 ()
  (solve-find-problem 'mshop-simple2 8 1 "sched-world"))
(defun ms5 ()
  (solve-find-problem 'mshop-simple5 4 1 "sched-world"))
(defun ms4 ()
  (solve-find-problem 'mshop-simple4 4 1 "sched-world"))
(defun mgri ()
  (solve-find-problem 'RAT-INSULIN-strips 8 1 "molgen"))
(defun mgmg1 ()
  (solve-find-problem 'MOLGEN1-STRIPS 8 1 "molgen"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Formula manipulation
;;;

(defvar test-formula
    '(:AND (:NOT (ON B B 0)) (:NOT (ON A B 0)) (:NOT (ON C B 0))
      (:NOT (ON B A 0))
      (:NOT (ON A A 0)) (:NOT (ON B C 0)) (:NOT (ON A C 0)) (:NOT (ON C C 0))
      (:NOT (CLEAR A 0)) (:NOT (HOLDING B 0)) (:NOT (HOLDING A 0))
      (:NOT (HOLDING C 0)) (:NOT (ON-TABLE C 0)) (OBJECT A 0) (OBJECT B 0)
      (OBJECT C 0) (ON-TABLE A 0) (ON-TABLE B 0) (ON C A 0) (CLEAR B 0)
      (CLEAR C 0) (ARM-EMPTY 0) (ON A B 2) (ON B C 2)
      (:OR (:NOT (UNSTACK C C 1)) (OBJECT C 0))
      (:OR (:NOT (UNSTACK C C 1)) (OBJECT C 0))
      (:OR (:NOT (UNSTACK C C 1)) (ON C C 0))
      (:OR (:NOT (UNSTACK C C 1)) (CLEAR C 0))
      (:OR (:NOT (UNSTACK C C 1)) (ARM-EMPTY 0))
      (:OR (:NOT (UNSTACK C C 1)) (:NOT (ON C C 2)))
      (:OR (:NOT (UNSTACK C C 1)) (:NOT (ARM-EMPTY 2)))
      (:OR (:NOT (UNSTACK C C 1)) (:NOT (CLEAR C 2)))
      (:OR (:NOT (UNSTACK C C 1)) (CLEAR C 2))
      (:OR (:NOT (UNSTACK C C 1)) (HOLDING C 2))
      (:OR (:NOT (UNSTACK C A 1)) (OBJECT C 0))
      (:OR (:NOT (UNSTACK C A 1)) (OBJECT A 0))
      (:OR (:NOT (UNSTACK C A 1)) (ON C A 0)))
  "A test formula for use in debugging CNF manipluation functions.")

;; Not used any more.
;; (defun test-formula-literals ()
;;   (formula-literals test-formula))

;; (defun test-numberize-formula ()
;;   (multiple-value-bind (ls ht) (numberize-formula
;; 				(lpropify-formula test-formula))
;;     (list ls ht)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Multi-solve simple
;;;

;; Note by default no simple frames

;; 1100-3000: DSx   done 1-9, files moved.
(defun ms-ds (n)
  (multi-solve-problem (intern (format nil "D1S1-~d" n)) "D1S1"
			    n n (+ 1000 (* 100 n))
			    *frame-strategies*
			    *action-representations*  ; action representations
			    *exclusion-strategies* ; maybe not parallel
			    '(no-types)  ; types
			    ))

;; 3100: Sussman (gentian, with many restarts)
(defun ms-suss ()
  (multi-solve-problem 'PRODIGY-SUSSMAN-3 "sussman"
			   6 6 3100))
(defun ms-suss-types ()			; done, files moved
  (multi-solve-types 'PRODIGY-SUSSMAN-3 "sussman" 6 3160 6))
(defun ms-suss-excl ()			; died on thistle
  (multi-solve-excl 'PRODIGY-SUSSMAN-3 "sussman" 6 3170 6))

;; 3200: Ferry (larkspur)
(defun ms-ferry ()
  (multi-solve-problem 'ferry7 "ferry-domain" 7 7 3200))

(defun ms-ferry3 ()
  (multi-solve-problem 'ferry3 "ferry-domain" 3 3 3210))

(defun ms-ferry-type ()			; done, files moved
  (multi-solve-types 'ferry7 "ferry-domain" 7 3260 7))
;; Perhaps don't bother; we know binary sucks.
(defun ms-ferry-excl ()			; trillium
  (multi-solve-excl 'ferry7 "ferry-domain" 7 3270 7))
;; Simple splitting  is failing.  UGH!
(defun ms-ferry-split ()
  (multi-solve-act-reps 'ferry7 "ferry-domain" 7 3280 7))


;; ; 3300-3400: Fridge.  These are too big, I think
;; (defun ms-fridgea ()
;;   (multi-solve-problem 'fixa-strips "fridge" 7 7 3300))
;; (defun ms-fridgeb ()
;;   (multi-solve-problem 'fixb-strips "fridge" 13 13 3400))

;; 3500-4000: Tires (5: columbine, 4: anemone, 2: buttercup)
(defun ms-tire1 ()
  (multi-solve-problem 'FIX1-strips "tire" 5 5 3500))
(defun ms-tire1-types ()		; died repeatedly on thistle
  (multi-solve-types 'FIX1-strips "tire" 5 3560 5))
(defun ms-tire2 ()
  (multi-solve-problem 'FIX2-strips "tire" 3 3 3600))
(defun ms-tire2-types ()		; shootingstar
  (multi-solve-types 'FIX2-strips "tire" 3 3660 3))
(defun ms-tire3 ()
  (multi-solve-problem 'FIX3-strips "tire" 6 6 3700))
(defun ms-tire3-types ()		; died repeatedly on larkspur
  (multi-solve-types 'FIX3-strips "tire" 6 3760 6))
(defun ms-tire4 ()
  (multi-solve-problem 'FIX4-strips "tire" 4 4 3800))
(defun ms-tire4-types ()		; died on dandelion
  (multi-solve-types 'FIX4-strips "tire" 4 3860 4))
(defun ms-tire5 ()
  (multi-solve-problem 'FIX5-strips "tire" 1 1 3900))
(defun ms-tire5-types ()		; died on buttercup.  What's up??
  (multi-solve-types 'FIX5-strips "tire" 1 3960 1))
;; Don't bother; this won't solve fast enough
(defun ms-tire ()
  (multi-solve-problem 'FIXIT-strips "tire" 19 19 4000))
(defun ms-tire-types ()
  (multi-solve-types 'FIXIT-strips "tire" 19 4060 19))

;; ;; 4100-4300: Monkeys.  These might be too big; all 3 failed.
;; (defun ms-monkey1 ()
;;   (multi-solve-find-problem 'MONKEY-TEST1 "monkey" X 4100 X))
;; (defun ms-monkey2 ()
;;   (multi-solve-find-problem 'MONKEY-TEST2 "monkey" X 4200 X))
;; (defun ms-monkey3 ()
;;   (multi-solve-find-problem 'MONKEY-TEST3 "monkey" X 4300 X))



;; 10000: sussman ("partial" style)
(defun ms-sussp-types ()		; died on shootingstar
  (multi-solve-types 'SUSSMAN-ANOMALY "blocks-world" 3 10060 3))
(defun ms-sussp-excl ()			; lupine
  (multi-solve-types 'SUSSMAN-ANOMALY "blocks-world" 3 10070 3))

;; 10200-10400: BLOCKS-WORLDn
(defun ms-bw1-types ()			; done, files moved
  (multi-solve-types 'simple-block1 "blocks-world" 3 10260 3))
(defun ms-bw1-excl ()			; died on saxifrage
  (multi-solve-excl 'simple-block1 "blocks-world" 3 10270 3))

(defun ms-bw2-types ()			; done, files moved
  (multi-solve-types 'simple-block2 "blocks-world" 4 10360 4))
(defun ms-bw2-excl ()			; shootingstar (will probably die)
  (multi-solve-excl 'simple-block2 "blocks-world" 4 10370 4))

;; 10500: hanoi1
(defun ms-hanoi1-types ()		; done, files moved
  (multi-solve-types 'HANOI-strips1 "hanoi" 1 10560 1))
(defun ms-hanoi1-excl ()		; died on thistle
  (multi-solve-excl 'HANOI-strips1 "hanoi" 1 10570 1))

;; 10600-10700: machine shop
(defun ms-ms5-types ()			; buttercup
  (multi-solve-types 'mSHOP-simple5 "sched-world" 2 10660 2))
(defun ms-ms4-types ()			; saxifrage
  (multi-solve-types 'mshop-simple4 "sched-world" 4 10760 4))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; File I/O
;;;

(defun test-read-tableau-file ()
  (list
   (read-tableau-file "~mernst/class/573/project/master/tests/PRODIGY-BW.PRODIGY-SUSSMAN.lisp-5-tb.out")
   (read-tableau-file "~mernst/class/573/project/master/tests/PRODIGY-BW.PRODIGY-SUSSMAN.lisp-6-tb.out")
   (read-tableau-file "~mernst/class/573/project/master/tests/D1S1.D1S1-1.lisp-0-tb.out")
   (read-tableau-file "~mernst/class/573/project/master/tests/D1S1.D1S1-3.lisp-2-tb.out")
   (read-tableau-file "~mernst/class/573/project/master/tests/D1S1.D1S1-5.lisp-4-tb.out")
))

(defun test-read-walksat-file ()
  (list
   (read-walksat-file "~mernst/class/573/project/master/tests/BLOCKS-WORLD-DOMAIN.SUSSMAN-ANOMALY.lisp-3-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-1.lisp-0-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-1.lisp-1-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-3.lisp-2-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-3.lisp-3-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-5.lisp-4-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/D1S1.D1S1-5.lisp-5-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/PRODIGY-BW.PRODIGY-SUSSMAN.lisp-5-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/PRODIGY-BW.PRODIGY-SUSSMAN.lisp-6-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/tiny-when.lisp-1-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/tiny-when.lisp-4-ws.out")
   (read-walksat-file "~mernst/class/573/project/master/tests/tiny.lisp-1-ws.out")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;		The *plan-length-table* creation and functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *plan-length-table*
    '(
      ;;; AT&T problems
      ;; Our smaller versions of AT&T problems
      (att-logistics0 3 3)
      (att-logistics1 6 6)
      (att-logistics2 9 9)
      (att-logistics2.5 10 9)
      (att-logistics3 10 9)
      ;; AT&T problems in paper
      (att-logistics-A 11 11)
      (att-sussman 3 3)
      (att-bw-large-a 6 6)
      (att-bw2-large-a 6 4)

      ;;; Non-AT&T problems
      (small-bw1 5 4)
      (med-bw1 7 5)
      (small-bw2 4 4)
      (med-bw2 5 5)
      (big-bw2 7 7)
      (big-bw1 11 7)

      (hanoi-strips1 1 1)
      (hanoi-strips2 3 3)
      (hanoi-strips3 7 7)
      (ferry3 3 3)
      (ferry7 7 7)
      (fix-strips1 9 7)
      (fix-strips2 14 10)

      (monkey-test1 6 6)
      (monkey-test2 9 9)
      (monkey-test3 10 10)

      (sussman1 3 3)
      (sussman2 3 3)

      (FIXA-strips 7 7)
      (FIXB-strips 13 13)

      ;;; Artificial domains
      (three-diamonds 10 4)
      (D1S1-1 1 1)
      (D1S1-2 2 2)
      (D1S1-3 3 3)
      (D1S1-4 4 4)
      (D1S1-5 5 5)
      (D1S1-6 6 6)
      (D1S1-7 7 7)
      (D1S1-8 8 8)
      (D1S1-9 9 9)
      (D1S1-10 10 10)
      (art1a 1 1)
      (art2a 2 2)
      (art3a 3 3)
      (art4a 4 4)
      (art5a 5 5)
      (art6a 6 6)
      (art1b 1 1)
      (art2b 2 2)
      (art3b 3 3)
      (art4b 4 4)
      (art5b 5 5)
      (art6b 6 6)
      (ART-1D-RD-DOMAIN1 1 1)
      (ART-1D-RD-DOMAIN2 1 1)
      (ART-1D-RD-DOMAIN3 1 1)
      (ART-1D-RD-DOMAIN4 1 1)
      (ART-1D-RD-DOMAIN5 1 1)
      (ART-1D-RD-DOMAIN6 2 2)
      (ART-1D-RD-DOMAIN7 2 2)
      (ART-1D-RD-DOMAIN8 2 2)
      (ART-1D-RD-DOMAIN9 3 3)
      (ART-1D-RD-DOMAIN10 3 3)
      (ART-1D-RD-DOMAIN11 3 3)
      (ART-1D-RD-DOMAIN12 3 3)
      (ART-1D-RD-DOMAIN13 3 3)
      (ART-1D-RD-DOMAIN14 3 3)
      (ART-1D-RD-DOMAIN15 3 3)
      (ART-1D-RD-DOMAIN16 5 5)
      (ART-1D-RD-DOMAIN17 4 4)
      (ART-1D-RD-DOMAIN18 5 5)
      (ART-1D-RD-DOMAIN19 4 4)
      (ART-1D-RD-DOMAIN20 5 5)
      (ART-1D-domain1 1 1)
      (ART-1D-domain2 1 1)
      (ART-1D-domain3 1 1)
      (ART-1D-domain4 1 1)
      (ART-1D-domain5 1 1)
      (ART-1D-domain6 2 2)
      (ART-1D-domain7 2 2)
      (ART-1D-domain8 2 2)
      (ART-1D-domain9 2 1)
      (ART-1D-domain10 2 2)
      (ART-1D-domain11 3 3)
      (ART-1D-domain12 3 3)
      (ART-1D-domain13 3 3)
      (ART-1D-domain14 3 2)
      (ART-1D-domain15 3 1)
      (ART-1D-domain16 4 2)
      (ART-1D-domain17 4 2)
      (ART-1D-domain18 4 2)
      (ART-1D-domain19 4 2)
      (ART-1D-domain20 4 2)
      (ART-MD-domain1 1 1)
      (ART-MD-domain2 1 1)
      (ART-MD-domain3 1 1)
      (ART-MD-domain4 1 1)
      (ART-MD-domain5 1 1)
      (ART-MD-domain6 2 2)
      (ART-MD-domain7 2 2)
      (ART-MD-domain8 2 2)
      (ART-MD-domain9 2 2)
      (ART-MD-domain10 2 2)
      (ART-MD-domain11 3 3)
      (ART-MD-domain12 3 3)
      (ART-MD-domain13 3 3)
      (ART-MD-domain14 3 3)
      (ART-MD-domain15 3 3)
      (ART-MD-domain16 4 4)
      (ART-MD-domain17 4 4)
      (ART-MD-domain18 4 4)
      (ART-MD-domain19 4 4)
      (ART-MD-domain20 4 4)
      (ART-MD-NS-domain1 1 1)
      (ART-MD-NS-domain2 1 1)
      (ART-MD-NS-domain3 1 1)
      (ART-MD-NS-domain4 1 1)
      (ART-MD-NS-domain5 1 1)
      (ART-MD-NS-domain6 2 2)
      (ART-MD-NS-domain7 2 2)
      (ART-MD-NS-domain8 2 2)
      (ART-MD-NS-domain9 2 2)
      (ART-MD-NS-domain10 2 2)
      (ART-MD-NS-domain11 3 3)
      (ART-MD-NS-domain12 3 3)
      (ART-MD-NS-domain13 3 3)
      (ART-MD-NS-domain14 3 3)
      (ART-MD-NS-domain15 3 3)
      (ART-MD-NS-domain16 4 4)
      (ART-MD-NS-domain17 4 4)
      (ART-MD-NS-domain18 4 4)
      (ART-MD-NS-domain19 4 4)
      (ART-MD-NS-domain20 4 4)
      (ART-MD-RD-domain1 1 1)
      (ART-MD-RD-domain2 1 1)
      (ART-MD-RD-domain3 1 1)
      (ART-MD-RD-domain4 1 1)
      (ART-MD-RD-domain5 1 1)
      (ART-MD-RD-domain6 3 3)
      (ART-MD-RD-domain7 3 3)
      (ART-MD-RD-domain8 2 2)
      (ART-MD-RD-domain9 3 3)
      (ART-MD-RD-domain10 2 2)
      (ART-MD-RD-domain11 3 3)
      (ART-MD-RD-domain12 4 4)
      (ART-MD-RD-domain13 4 4)
      (ART-MD-RD-domain14 4 4)
      (ART-MD-RD-domain15 5 5)
      (ART-MD-RD-domain16 5 5)
      (ART-MD-RD-domain17 4 4)
      (ART-MD-RD-domain18 6 6)
      (ART-MD-RD-domain19 6 6)
      (ART-MD-RD-domain20 5 5)
      (ART-MD-NS-RD-domain1 1 1)
      (ART-MD-NS-RD-domain2 1 1)
      (ART-MD-NS-RD-domain3 1 1)
      (ART-MD-NS-RD-domain4 1 1)
      (ART-MD-NS-RD-domain5 1 1)
      (ART-MD-NS-RD-domain6 3 3)
      (ART-MD-NS-RD-domain7 3 3)
      (ART-MD-NS-RD-domain8 2 2)
      (ART-MD-NS-RD-domain9 2 2)
      (ART-MD-NS-RD-domain10 2 2)
      (ART-MD-NS-RD-domain11 3 3)
      (ART-MD-NS-RD-domain12 4 4)
      (ART-MD-NS-RD-domain13 4 4)
      (ART-MD-NS-RD-domain14 3 3)
      (ART-MD-NS-RD-domain15 3 3)
      (ART-MD-NS-RD-domain16 4 4)
      (ART-MD-NS-RD-domain17 6 6)
      (ART-MD-NS-RD-domain18 4 4)
      (ART-MD-NS-RD-domain19 4 4)
      (ART-MD-NS-RD-domain20 6 6)
      )
  "List of lists of (basename sequential-length parallel-length) for
minimal-length plans solving the given problem.")



;; Given a filename (just the basename), this function returns the
;; appropriate plan length to use, depending on whether the current
;; flags represent a sequential or a parallel encoding.
;;;(defun get-plan-length (probname)
;;;  (let ((lookup (cdr (assoc probname *plan-length-table*))))
;;;    (cond ((null lookup)
;;;	   (error "Problem ~a is not in the *plan-length-table*" probname))
;;;	  ((eq *exclusion-strategy* 'parallel)
;;;	   (cadr lookup))
;;;	  (t (car lookup)))))

;; Given a filename (just the basename), this function returns the
;; appropriate plan length to use, depending on whether the current
;; flags represent a sequential or a parallel encoding.
(defun get-plan-length (probname &optional no-error)
  (let ((length-alist (problem-length (get-problem probname))))
    (cond ((eq *exclusion-strategy* 'parallel)
	   (second (assoc :parallel length-alist)))
	  ((eq *exclusion-strategy* 'sequential-pairwise)
	   (second (assoc :serial length-alist)))
	  (no-error
	   nil)
	  (t
	   (error "Problem ~A does not define length information" probname)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mike's tracing examples
;;;

;; Consider inserting (break) at various points in the code and rerunning.

;; For CMU Common Lisp tracing output:
;; (setq debug:*debug-print-length* nil debug:*debug-print-level* nil)

;; (untrace)
;; (trace solve-problem find-plan write-cnf-file read-tableau-file read-walksat-file)
;; Probably a bad idea: too much output
;; ;; (trace instantiate-init-step-and-goal tableau-solve-cnf walksat-solve-cnf)

(defun trace-clauses ()
  (trace action-clauses exclusion-clauses frame-clauses init-action-and-goal-clauses))

;; ;; (trace numberize-formula formula-literals)
;; ;; (trace flatten-expr)

;; (step find-plan)

;; In order from most to least frequently called:
;; (trace unchanged-prop->simp-frame-clauses)
;; (trace act-prop->simp-frame-clauses)
;; (trace get-simp-frame-clauses get-act-prop-list)

;; (trace split-clauses-simple split-action-simple)

;; (trace prop-substitute prop-substitute-one)

;; (trace find-plan solve-cnf tableau-solve-cnf walksat-solve-cnf)

;; (trace distribute-ors-over-ands combinations-multiple)

;; (trace plan-odd-time-only)

;; (trace frame-clauses exclusion-clauses plan-odd-time-only factored-at-least-one-extra-vars partial-action-elim-extra-vars)

;; The most common functions to trace
;; (trace-clauses)

;; For generating .cnf files:
;; (setq *dont-solve* nil *external-simplifier* nil)
