(in-package medic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Logic conversion
;;;

;; ;; Should also check for t in disjuncts, in which case return t, and eliminate
;; ;; nil from disjuncts.  (Like binarize does.)
;; (defun make-disjunction (disjuncts)
;;   (let ((result (disjuntion disjuncts)))
;;     (if (eq result 'false)
;; 	nil
;;       result)))


;;; Formulas


(defun formula-p (formula)
  ;; was  (or (symbol-or-number-p formula) ...)
  (or (sprop-p formula)
      (and (listp formula)
	   (not (null formula))
	   (let ((fn (first formula)))
	     (cond ((eq :not fn)
		    (and (length-two formula)
			 (not (sprop-p (second formula)))
			 (formula-p (second formula))))
		   ;; ((eq 'implies fn)
		   ;; (and (= 3 (length formula))
		   ;;	 (formula-p (second formula))
		   ;;	 (formula-p (third formula))))
		   ((or (eq :and fn) (eq :or fn))
		    (and  ;;  (cdr formula)	; length > 1
		     (clauses-p (cdr formula))))
		   (t
		    nil))))))

;; Not used any more.
;; ;; List of literals.  May be inefficient.  Perhaps use hash table instead,
;; ;; if the formula is large enough.
;; (defun formula-literals (whole-formula)
;;   (let ((result '()))
;;     (labels ((process (formula)
;; 	       (if (symbol-or-number-p formula)
;; 		   (setq result (adjoin formula result))
;; 		 (let ((fn (first formula)))
;; 		   (cond ((eq :not fn)
;; 			  (process (second formula)))
;; 			 ;; ((eq 'implies fn)
;; 			 ;;  (process (second formula))
;; 			 ;;  (process (third formula)))
;; 			 ((or (eq :and fn) (eq :or fn))
;; 			  (dolist (subformula (cdr formula))
;; 			    (process subformula)))
;; 			 (t
;; 			  (setq result
;; 			    (adjoin formula result :test #'equal))))))))
;;       (process whole-formula)
;;       result)))

(defun clauses-p (l)
  (every #'formula-p l))


(defun make-implies (predicate consequence)
  ;; (list 'implies predicate consequence)
  ;; one level isn't enough
  ;; (list :or (push-negation-once predicate) consequence)
  (list :or (formula-negate predicate) consequence))

(defun conjunctive-clauses (formula)
  (if (and (listp formula)
	   (eq :and (first formula)))
      (cdr formula)
    (list formula)))

(defun disjunctive-clauses (formula)
  (if (and (listp formula)
	   (eq :or (first formula)))
      (cdr formula)
    (list formula)))

;; alternate name: conjunction-strip-and-maybe
(defun conjunction-simplify (formula)
  (if (and (listp formula)
	   (eq :and (first formula))
	   (length-one formula))
      (second formula)
    formula))

;;; No longer used; we no longer permit "implies".
;; ;; Convert (implies A B) to (or (not A) B)
;; (defun eliminate-implies (formula)
;;   (cond ((symbol-or-number-p formula)
;; 	 formula)
;; 	((and (listp formula)
;; 	      (eq 'implies (car formula))
;; 	      (= 3 (length formula)))
;; 	 `(or (not ,(eliminate-implies (second formula)))
;; 	      ,(eliminate-implies (third formula))))
;; 	(t
;; 	 (mapcar #'eliminate-implies formula))))
;; ;; (eliminate-implies '(and (implies (and b c) d) (implies (implies e f) g)))


;;; CNF formulas

;; Not used.
;; (defun convert-to-cnf-clauses (clauses)
;;   (cdr (convert-to-cnf (cons :and clauses))))

;; Returns a list of clauses:  if arg is conjunction, its operands, else list of arg.
(defun cnf-clauses (cnf-formula)
  (conjunctive-clauses cnf-formula))

;; Warning: this side-effects its arguments.
(defun cnf-conjoin (&rest cnf-formulas)
  (assert (every #'cnf-p cnf-formulas))
  (let ((result (cons :and (apply #'append (mapcar #'cnf-clauses cnf-formulas)))))
    (assert (cnf-p result))
    result))

;; like cnf-clauses, but works even if top-level formula consists of nested ands.
;; does the real work of flatten-ands, below.
(defun cnf-clauses-recurse (cnf-formula)
  (if (and (listp cnf-formula)
	   (eq :and (first cnf-formula)))
      (apply #'append (mapcar #'cnf-clauses-recurse
			      (cdr cnf-formula)))
    (list cnf-formula)))
;; Testing
;; (cnf-clauses-recurse '(:or a b c))
;; (cnf-clauses-recurse '(:and a b c))
;; (cnf-clauses-recurse '(:and a (:and b c)))
;; (cnf-clauses-recurse '(:and a (:or b c)))
;; (cnf-clauses-recurse '(:and a (:or (:and b c) d)))

(defun disjunctive-clauses-recurse (formula)
  (if (and (listp formula)
	   (eq :or (first formula)))
      (apply #'append (mapcar #'disjunctive-clauses-recurse
			      (cdr formula)))
    (list formula)))


(defun flatten-ands (formula)
  (let ((results (cnf-clauses-recurse formula)))
    (if (and results (length-one results))
	(car results)
      (cons :and results))))

(defun flatten-ors (formula)
  (let ((results (disjunctive-clauses-recurse formula)))
    (if (and results (length-one results))
	(car results)
      (cons :or results))))

(defun flatten-formula (formula)
  (if (not (consp formula))
      formula
    (let ((new-formula (mapcar #'flatten-formula formula)))
      (case (car new-formula)
	((:not)
	 (if (and (consp (second new-formula))
		  (eq :not (car (second new-formula))))
	     (second (second new-formula))
	   new-formula))
	((:and)
	 (flatten-ands new-formula))
	((:or)
	 (flatten-ors new-formula))
	(otherwise
	 new-formula)))))
;; Testing
;; (flatten-formula '(:OR (:AND (:NOT (ANNOYED :AFTER)))))
;; (flatten-formula '(:OR (:AND (:NOT (:NOT (ANNOYED :AFTER))))))

;;; CNF predicates

(defun disjunction-p (formula)
  (and (listp formula)
       (eq :or (car formula))))

(defun conjunction-p (formula)
  (and (listp formula)
       (eq :and (car formula))))

(defun cnf-p (formula)
  (or (and (conjunction-p formula)
	   (cnf-clauses-p (cdr formula)))
      (disjunction-of-propositions-p formula)))

(defun cnf-clauses-p (clauses)
  (every #'disjunction-of-propositions-p
	 clauses))

(defun disjunction-of-propositions-p (formula)
  ;; was (or (symbol-or-number-p formula) ...)
  (or (sprop-p formula)
      (and (listp formula)
	   (eq :or (first formula))
	   (every #'sprop-p (cdr formula)))))

;; Used to include :implies.  These are the connectives for logical formulae.
(defvar *formula-combiners* '(:and :or :not))
(defun formula-combiner-p (symbol)
  (member symbol *formula-combiners* :test #'eq))

;; Convert two-element list of (not sprop) into a negated sprop.
(defun cnf-fix-sprop-negation (cnf)
  (cond ((not (consp cnf))
	 cnf)
	((eq :and (car cnf))
	 (cons :and (mapcar #'disjunction-fix-sprop-negation (cdr cnf))))
	((eq :or (car cnf))
	 (disjunction-fix-sprop-negation cnf))
	((eq :not (car cnf))
	 (assert (sprop-p (second cnf)))
	 (sprop-negate (second cnf)))
	(t
	 (error "What CNF formula? (non-sprop?) ~s" cnf))))

;; Convert two-element list of (not sprop) into a negated sprop.
(defun disjunction-fix-sprop-negation (disjunction)
  (assert (and (consp disjunction) (eq :or (car disjunction))))
  (cons :or (mapcar #'sprop-fix-negation (cdr disjunction))))

;; Convert two-element list of (not sprop) into a negated sprop.
(defun sprop-fix-negation (x)
  (cond ((sprop-p x)
	 x)
	((and (consp x) (eq :not (car x)))
	 (sprop-negate (second x)))
	(t
	 (error "Bad argument to sprop-fix-negation: ~s" x))))


;; ;; An atomic predicate does not contain combiners like and, or, not
;; (defun atomic-predicate-p (formula &optional permit-negation)
;;   ;; was (or (symbol-or-number-p formula) ...)
;;   (and (listp formula)
;;        (let ((fn (car formula)))
;;          (if (eq :not fn)
;;              (and permit-negation
;;                   (length-two formula)
;;                   (atomic-predicate-p (second formula)))
;;            (and (not (formula-combiner-p fn))
;;                 (every #'symbol-or-number-p formula)
;;                 (notany #'formula-combiner-p formula))))))

(deftype symbol-or-number ()
  '(or symbol fixnum))
(deftype keyword-or-number ()
  '(or keyword fixnum))

(defun symbol-or-number-p (x)
  (or (symbolp x) (numberp x)))
(defun keyword-or-number-p (x)
  (or (keywordp x) (numberp x)))


;;; Test CNF formula for triviality

(defun cnf-true-or-false (bool)
  (if bool
      (cnf-true)
    (cnf-false)))

(defun cnf-true ()
  ;; not '(:and); we want a fresh copy for each
  (list :and))

;; tests for a true CNF formula
(defun cnf-true-p (x)
  (or
   ;; form-cnf
   (equal x '(:and t))
   ;; ->cnf
   (eq x ':true)
   (equal x '(:true))
   ;; general
   (equal x '(:and))))

(defun cnf-false ()
  ;; not '(:or); we want a fresh copy for each
  (list :or))

;; tests for a false CNF formula
(defun cnf-false-p (x)
  (or
   ;; form-cnf
   (equal x '(:and nil))
   ;; ->cnf
   (eq x :false)
   (equal x '(:false))
   (and (listp x) (member :false x))
   ;; dpll
   (equal x '(:and (:or)))
   ;; general
   (equal x '(:or))))


;; turn this formula into a trivially true formula
(defun make-formula-true (formula)
  (assert (consp formula))
  (setf (car formula) :and)
  (setf (cdr formula) nil))

;;; Distribute ors over ands

;; Takes a single disjunction, where each disjunct is a possibly negated
;;   atomic predicate or a conjunction or disjunction of such.
;;   (Maybe should permit possibly negated conj. or disj. of such.)
;; Returns a list of disjunctions, where each disjunct is a possibly
;;   negated atomic predicate -- ie, an element of a CNF formula.
(defun distribute-ors-over-ands (formula)
  (if (sprop-p formula)
      (list formula)
    (case (first formula)
      (:not
       (let ((negated (second formula)))
	 (cond ((sprop-p negated)
		(list (sprop-negate negated)))
	       ((eq :and (first negated))
		(list (cons :or (mapcar #'sprop-negate (cdr negated)))))
	       (t
		(error "What formula? ~s" formula)
		;; no change
		;; (list formula)
		))))
      (:and
       (cdr formula))
      (:or
       ;; each disjunct is either a conjunction or not
       (let ((conjunction-elts-lists '())
	     (non-conjunctions '()))
	 (declare (list conjunction-elts-lists) (list non-conjunctions))
	 (dolist (disjunct (cdr formula))
	   (if (sprop-p disjunct)
	       (push disjunct non-conjunctions)
	     (case (car disjunct)
	       (:and
		;; Replaced by the below; is it worth it?
		;;  (push (cdr disjunct) conjunction-elts-lists)
		(if (cdr disjunct)
		    ;; at least one conjunct
		    (if (cddr disjunct)
			;; at least two conjuncts
			(push (cdr disjunct) conjunction-elts-lists)
		      ;; just one conjunct
		      (push (second disjunct) non-conjunctions))
		  ;; no conjuncts; nothing to do
		  ))
	       (:or
		;; This copy-list is not strictly needed
		;; because of copy-tree in
		;; prop-substitute-one, but be safe.
		(dolist (sub-disjunct (cdr disjunct))
		  (assert (sprop-p sub-disjunct))
		  (push sub-disjunct non-conjunctions)))
	       ;; (not
	       ;; (assert (eq :and (first (second disjunct))))
	       ;; (dolist (negated-conjunct (cdr (second disjunct)))
	       ;;  (push (sprop-negate negated-conjunct) non-conjunctions)))
	       (t
		(error "What disjunct? ~s" disjunct)))))
	 ;; (format t "conj: ~s~%non-conj: ~s~%" conjunction-elts-lists non-conjunctions)
	 ;; gratuitous, but puts lists in more like their original order.
	 ;; (setq conjunction-elts-lists (nreverse conjunction-elts-lists))
	 ;; (setq non-conjunctions (nreverse non-conjunctions))

	 ;; Will the tableau simplifier do this?  Is it better for me to
	 ;; do it here anyway?
	 (setq non-conjunctions (delete-duplicates non-conjunctions :test #'sprop-equal))
	 ;; assert function contained (format t "~&duplicate check: ~s~%" x)
	 (assert (notany #'(lambda (x) (clause-contains-duplicate-sprop (cons :or x))) conjunction-elts-lists))
	 ;; (format t "~&comb-mult1 ~s~%comb-mult2 ~s~%" conjunction-elts-lists non-conjunctions)

	 (let ((max-combinations (apply #'* (mapcar #'length conjunction-elts-lists))))
	   (declare (integer max-combinations))
	   ;; (if (> max-combinations 10000000) ; 10 million
	   ;;     (error "combinations-multiple trying to build ~d combinations" max-combinations))
	   (if (> max-combinations 10000) ; 10 thousand
	       (format t "~&WARNING: combinations-multiple was asked to build ~d combinations,~%will try to build fewer if possible (as is usually the case), from~%  ~S~%  ~S~% originally ~S~%" max-combinations conjunction-elts-lists non-conjunctions formula))

	   ;; Further optimizations; repeat until settled:
	   ;;  * if anything is in every clause (their intersection is
	   ;;    non-empty), then remove it from each and adjoin it to the init,
	   ;;    and remove any clauses that are now empty.
	   ;;  * if any clause has length one, then adjoin it to the init
	   ;;    and remove that clause (but don't remove that prop from other
	   ;;    clauses).
	   (let* ((non-duplicate-clause-contents
		   (combinations-multiple-no-repeat
		     conjunction-elts-lists
		     :initial-value non-conjunctions))
		  (non-dup-non-neg-clause-contents
		   (remove-if #'(lambda (cl)
				  (clause-contains-sprop-and-negation (cons :or cl)))
			      non-duplicate-clause-contents))
		  ;; If I have both A v B v C and A v B, then kill A v B v C.
		  ;; Perhaps I could have done this earlier, but then
		  ;; combinations-multiple-no-repeat would be less general.
		  (non-subsumed-clause-contents
		   (remove-if #'(lambda (cl)
				  (some #'(lambda (smaller)
					    (strict-subset-p smaller cl
							     :test #'sprop-equal))
					non-dup-non-neg-clause-contents))
			      non-dup-non-neg-clause-contents))
		  (result
		   (mapcar
		    #'(lambda (disjuncts)
			;; remove-duplicates is unnecesary if postprocessor
			;; removes duplicates; combinations-multiple-no-repeat
			;; shouldn't create duplicates
			;; (cons :or (remove-duplicates disjuncts :test #'sprop-equal))
			(assert (not (clause-contains-duplicate-sprop (cons :or disjuncts))))
			(cons :or disjuncts))
		    non-subsumed-clause-contents)))
	     (declare (list non-duplicate-clause-contents)
		      (list non-dup-non-neg-clause-contents)
		      (list non-subsumed-clause-contents)
		      (list result))
	     (if (> max-combinations 10000) ; 10 thousand
		 (format t "~&combinations-multiple built ~d clauses (~d before subsumption), not ~d~%"
			 (length result) (length non-duplicate-clause-contents) max-combinations))
	     result))))
      (otherwise
       (error "What formula? ~s" formula)))))
;; (distribute-ors-over-ands '(:or (a) (b) (c)))
;; (distribute-ors-over-ands '(:or (a) (:and (b) (c))))
;; (distribute-ors-over-ands '(:or (a) (:and (b) (c)) (d)))
;; (distribute-ors-over-ands '(:or (a) (:and (b) (c)) (:and (d) (e) (f)) (g)))
;; (distribute-ors-over-ands '(:or (:and (b) (c)) (:and (d) (e)) (:and (f) (g))))


;;; Formula size

;; Returns a list of three values: number of distinct variables (negation
;; doesn't add to this number), number of clauses, number of literals (=
;; sum of clause sizes).
(defun formula-size (formula)
  (declare (list formula))
  (let ((table (formula-occurrences formula)))
    (list (hash-table-count table)
	  (if (formula-combiner-p (car formula))
	      ;; more efficient than (length (cdr formula)) becuse of type info
	      (1- (length formula))
	    1)
	  (let ((result 0))
	    (declare (fixnum result))
	    (maphash #'(lambda (key val)
			 (declare (ignore key))
			 (declare (fixnum val))
			 (incf result val))
		     table)
	    result))))

;; Returns a hash table from unnegated atomic propositions to the number of
;; times each appears in the formula.
(defun formula-occurrences (formula)
  (let ((table (make-sprop-hash-table)))
    (labels ((process (f)
	       (cond ((sprop-p f)
		      (incf (gethash (sprop-nonnegated f) table 0)))
		     ((formula-combiner-p (car f))
		      (dolist (sub-f (cdr f))
			(process sub-f)))
		     (t
		      (error "What f? ~s" f)))))
      (process formula))
    table))

(defun formula-occurrences->histogram (table)
  ;; key = # of occurences, val = # of vars ocurring that often
  (let ((occur-table (make-hash-table)))
    (maphash #'(lambda (key val)
		 (declare (ignore key))
		 (incf (gethash val occur-table 0)))
	     table)
    (let ((result '()))
      (maphash #'(lambda (key val)
		   (push (list key val) result))
	       occur-table)
      (sort result #'< :key #'first))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Running programs
;;;

(defvar *time-command* "time")
(defvar *time-command-extra-args* nil
  "List of extra arguments to 'time' Unix command.
Possible values:  nil, '(\"nice\"), '(\"polite\").")

(defvar *run-program-timeout* nil
  "If non-nil, then maximum CPU usage (in seconds) permitted to run-program.")


;; In CMUCL this would call the "time" program rather than the built-in
;; shell command, which I would prefer.  So at present this function does
;; no timing, just calls run-program.
;; I probably want to supply args ":error-output :output" so I see the
;; result of the time program.
(defun time-program (program program-args &rest run-program-args)
  ;;   (apply #'run-program "time"
  ;; 	 (append *time-command-extra-args* (cons program program-args))
  ;; 	 run-program-args)
  (apply #'run-program program program-args
	 run-program-args))

;; Problem: in ACL, :output nil means use standard output
;; while in CMUCL, :output t means use standard and nil means use /dev/null.
(defun run-program (program program-args
		    &key
		    ;; common to ACL and CMUCL
		    input output (wait t) (if-input-does-not-exist :error) (if-output-exists :error)
		    ;; must be translated for CMUCL
		    error-output (if-error-output-exists :error)
		    ;; only CMUCL; before-execve is documented but not supported
		    (env #+:allegro nil #+:cmu ext:*environment-list*)
		    pty status-hook)
  "Run an external program.  Works under CMUCL and ACL.
PROGRAM-ARGS shouldn't depend on whether globbing is performed;
and they will not be globbed.
Keyword arguments:
 As in both CMUCL's run-program and ACL's run-shell-command:
   :input :output :wait :if-input-does-not-exist :if-output-exists
 As in ACL; corresponds to CMUCL's :error, :if-error-exists:
   :error-output :if-error-output-exists
 Only meaningful in CMUCL:
   :env :pty :status-hook
Defaults for the :if-* keyword arguments are :error, as in ACL but not CMUCL."

  ;; Perhaps not the most elegant way to implement this functionality; oh well.
  (if *run-program-timeout*
      (setq program-args (cons (format nil "~a" *run-program-timeout*)
			       (cons program program-args))
	    program "exec_cpu_limited"))

  ;; CMU CL does no globbing of arguments; "ls" '("-l" "*.lisp") :output t
  ;; gives back "ls: *.lisp: No such file or directory".
  #+:cmu
  (let ((process (funcall #'extensions:run-program program program-args
			  :input input :output output :wait wait
			  :if-input-does-not-exist if-input-does-not-exist
			  :if-output-exists if-output-exists
 			  :error error-output
 			  :if-error-exists if-error-output-exists
			  :env env :pty pty :status-hook status-hook)))
    ;; Should really call process-close when done
    (ext:process-output process))

  ;; ACL hands the whole command line to the shell, which then does globbing.
  ;; This could be made to deal with env; the others are problematic
  #+:allegro (if (or env pty status-hook)
		 (error "Can't specify env (~s), pty (~s), or status-hook (~s) in Allegro Common Lisp" env pty status-hook)
	       (funcall #'excl:run-shell-command
			(format nil "~a~{ '~a'~}" program program-args)
			:input input :output output :wait wait
			:if-input-does-not-exist if-input-does-not-exist
			:if-output-exists if-output-exists
			:error-output error-output
			:if-error-output-exists if-error-output-exists)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Other utilities
;;;

(deftype bool ()
  "t or nil"
  '(member t nil))

;; Return a list containing the first n elements of the given list.
(defun firstn (n list)
  (declare (fixnum n) (list list))
  (butlast list (- (length list) n)))
;; Testing
;; (firstn 3 '(a b c d e))
;; (firstn 7 '(a b c d e))
;; (firstn 5 '(a b c d e))

(defun xor (a b)
  (or (and (not a) b)
      (and a (not b))))
;; (xor nil nil)
;; (xor nil t)
;; (xor t nil)
;; (xor t t)

;; an infinite argument version of union
(defun union-of-lists (&rest args)
  (reduce #'union args))


;; Maybe return the empty list, for the purposes of #'reduce.
(defun union-equal (a b)
  (union a b :test #'equal))
;; (declaim (inline union-equal))

(defun intersection-equal (a b)
  (intersection a b :test #'equal))
;; (declaim (inline intersection-equal))

(defun set-difference-equal (a b)
  (set-difference a b :test #'equal))
;; (declaim (inline set-difference-equal))

(defun member-equal (a b)
  (member a b :test #'equal))
;; (declaim (inline member-equal))

;; Use #'some instead; it's more efficient.
;; (defun reduce-or (l)
;;   (reduce #'(lambda (a b) (or a b))
;; 	  l))

;; Recursively find the first subpart of l satisfying function f.
;; Return nil if no such thing is found.
(defun rec-find-if (f l)
  (declare (type (function (*) *) f))
  (if (funcall f l)
      l
    (and (listp l)
	 (some #'(lambda (x)
		   (rec-find-if f x))
	       l))
    ;;     (some #'(lambda (x)
    ;;               (and (listp x) (rec-find-if f x)))
    ;;           l)
    ))

;; (defun rec-find-if-throttled (f l recurse-fn)
;;   (if (funcall f l)
;;       l
;;     (and (listp l)
;; 	 (funcall recurse-fn l)
;; 	 (some #'(lambda (x)
;; 		   (and (listp x) (rec-find-if-throttled f x recurse-fn)))
;; 	       l))))

;; recursive remove-if
;; only works if function F checks a property of lists
;; WARNING: this does the copy before the test, so beware use of EQ
(defun rec-remove-if (f l)
  (if (atom l)
      l
    (remove-if f (mapcar #'(lambda (e)
			     (rec-remove-if f e))
			 l))))

;; here's a version of remove-if-not for retrieving all atoms satisfying
;; a certain property, from an arbitrarily nested list
(defun rec-retrieve-atom-if (f l)
  (multiple-value-bind (lsts atms) (split #'listp l)
    (let ((pass (remove-if-not f atms)))
      (if (null lsts) pass
	(union pass (reduce #'union
			    (mapcar #'(lambda (e)
					(rec-retrieve-atom-if f e))
				    lsts)))))))


;; ;; If recurse-fn returns t, do substitution inside the form.
;; ;; If recurse-fn returns nil, or it's an atom, don't recurse.
;; (defun rec-remove-if-throttled (f l recurse-fn)
;;   (if (or (atom l)
;;           (not (funcall recurse-fn l)))
;;       l
;;     (remove-if f (mapcar #'(lambda (e)
;;                              (rec-remove-if-throttled f e recurse-fn))
;;                          l))))


;; Given a number and a list of objects, make all possible num-long
;; combinations, with repetition; order is significant.
;; warning: list structure is shared.
(defun combinations (num objs)
  (declare (fixnum num))
  (cond ((= 0 num) ())
	((= 1 num) (mapcar #'list objs))
	(t (let ((induct-step (combinations (- num 1) objs)))
	     (mapcan #'(lambda (c)
			 (mapcar #'(lambda (o) (cons o c))
				 objs))
		     induct-step)))))
;; Testing
;; (combinations 2 '(a b))
;; (combinations 3 '(a b c))

;; Choose NUM of the obs, without repetition.  Order is irrelevant.
;; Returns a list of the results.
(defun choose (num objs)
  (declare (fixnum num) (list objs))
  (cond ((= num 0)
	 '())
	((= num 1)
	 (mapcar #'list objs))
	((> num (length objs))
	 (error "Can't choose ~d objects from shorter list ~s" num objs))
	((= num (length objs))
	 (list objs))
	(t
	 (nconc (choose num (cdr objs))
		(mapcar #'(lambda (sub-choice)
			    (cons (car objs) sub-choice))
			(choose (1- num) (cdr objs)))))))
;; Testing
;; (choose 3 '(a b))
;; (choose 3 '(a b c))
;; (choose 3 '(a b c d))


;; Pad the list by adding FILL (default nil) to end to make the result GOAL-LEN
;; elements long.  Does not truncate if original argument is too long.
;; This is useful for padding argumentst to mapcar, which stops after shortest
;; argument runs out.
(defun pad-list (lst goal-len &optional fill)
  (declare (list lst) (fixnum goal-len))
  (let ((lst-len (length lst)))
    (if (> goal-len lst-len)
	(append lst (make-list (- goal-len lst-len) :initial-element fill))
      lst)))
;; Testing
;; (pad-list '(a b c) 5)
;; (pad-list '(a b c) 3)
;; (pad-list '(a b c) 2)
;; (pad-list '(a b c) 5 'foo)

;; Return the list of every cons of an element of list1 onto an element of list2.
;; Result shares list structure.
(defun multi-cons (list1 list2)
  (mapcan #'(lambda (elt1)
	      (mapcar #'(lambda (elt2)
			  (cons elt1 elt2))
		      list2))
	  list1))

;; All combinations which consist of one element from each of the lists.
;; keyword argument :initial-value is a list whose elements all appear
;; literally in every resulting combination.
(defun combinations-multiple (lists &key (initial-value '()))
  ;; (format t "comb-mult ~s~%" lists)
  (reduce #'multi-cons lists :from-end t :initial-value (list initial-value)))
;; (combinations-multiple '((a b c) (d e)))
;; (combinations-multiple '((a b c) (d e) (last)))
;; (combinations-multiple '((a b c) (d e)) :initial-value '(last two))



;; Assumes no repeated elements in lists.
(defun set-equal-p (list1 list2 &key (test #'eql))
  (declare (list list1) (list list2))
  (and (= (length list1) (length list2))
       ;; (= (length list1) (length (intersection list1 list2 :test test)))
       (subsetp list1 list2 :test test)
       (subsetp list2 list1 :test test)))

;; Assumes no repeated elements in lists.
(defun strict-subset-p (list1 list2 &key (test #'eql))
  (declare (list list1) (list list2))
  (and (< (length list1) (length list2))
       (subsetp list1 list2 :test test)))


;; Return the list of every adjoin of an element of list1 onto an element
;; of list2.
;; Result shares list structure.
(defun multi-adjoin (list1 list2 &key (test #'eql))
  (let ((with-duplicates (mapcan #'(lambda (elt1)
				     (mapcar #'(lambda (elt2)
						 (adjoin elt1 elt2 :test test))
					     list2))
				 list1)))
    ;; Is this way too expensive?  Should I keep the lists ordered and
    ;; compare element-by-element?
    (remove-duplicates with-duplicates
		       :test #'(lambda (x y)
				 (set-equal-p x y :test test)))))

(defun combinations-multiple-no-repeat (lists &key (initial-value '()) (test #'eql))
  ;; (format t "comb-mult ~s~%" lists)
  (reduce #'(lambda (l1 l2)
	      (multi-adjoin l1 l2 :test test))
	  lists
	  :from-end t :initial-value (list initial-value)))
;; (combinations-multiple-no-repeat '((a b c) (d e)))
;; (combinations-multiple-no-repeat '((a b c) (d e) (last)))
;; (combinations-multiple-no-repeat '((a b c) (d e)) :initial-value '(last two))
;; (combinations-multiple-no-repeat '((a b) (b c)))
;; (combinations-multiple-no-repeat '((a b) (b c)) :initial-value '(c d e))
;; (combinations-multiple-no-repeat '((a c) (a d) (b c) (b d))) ; len 7, not 16
;; (combinations-multiple-no-repeat '((a b e) (a b f) (a b g) (a c e) (a c f) (a c g) (a d e) (a d f) (a d g)))  ;; len 79, not 19683


;; Assumes OBJS contains no two elements eql to one another (they are
;; considered to appear just once).
;; There must be a better way of doing this!
(defun permutations (objs)
  (if (null objs)
      '(())
    (mapcan #'(lambda (obj)
		(mapcar #'(lambda (subpermutation)
			    (cons obj subpermutation))
			(permutations (remove obj objs))))
	    objs)))

(defun random-permutation (objs)
  (if (null objs)
      '()
    (let ((elt (random-element objs)))
      (cons elt (random-permutation (remove elt objs))))))
;; (random-permutation '(a b c d e f))

(defun random-element (objs)
  (nth (random (length objs)) objs))
;; (random-element '(a b c d e f))


;; Return a new list just like LIST except its ELTNOth element is NEWELT
;; If ELTNO is greater than the length of list, NEWELT goes at the end.
;; Can do by side effect with (setf (nth ELTNO LIST) NEWELT)
(defun splice-at (eltno newelt list)
  (declare (fixnum eltno) (list list))
  (let ((listlen (length list)))
    (append (butlast list (- listlen eltno))
	    (list newelt)
	    (nthcdr (+ eltno 1) list))))

;; Old, inefficient, unpleasant implementation.
;; (defun contains-duplicate (arg &key (test #'eql))
;;   (let ((len (length arg)))
;;     (not (= len (length (remove-duplicates arg :test test))))))

;; Should I map with key first?  That might prevent extra calls to key...
(defun contains-duplicate (arg &key (test #'eql))
  (labels ((contains-duplicate-helper (arg)
	     (and (not (length-one arg))
		  (or (member (car arg) (cdr arg) :test test)
		      (contains-duplicate-helper (cdr arg))))))
    (and (not (null arg))
	 (contains-duplicate-helper arg))))

;; Returns all members that are duplicates, not just one copy of each; the
;; result will contain members for which TEST returns true.
(defun duplicates (arg &key (test #'eql))
  (declare (list arg) (type (function (* *) *) test))
  (let ((result '()))
    (labels ((duplicates-helper (lst)
	       ;; side-effects its argument
	       (if (not (or (null lst)
			    (length-one lst)))
		   (let* ((this-elt (car lst))
			  (found-duplicate nil)
			  (sans-first (cdr lst)) ; so we can declare its type
			  (this-recursive-arg
			   (delete-if
			    #'(lambda (elt)
				(if (funcall test this-elt elt)
				    (progn (push elt result)
					   (if (not found-duplicate)
					       (progn
						 (push this-elt result)
						 (setq found-duplicate t)
						 t)))))
			    sans-first)))
		     (declare (list sans-first))
		     (duplicates-helper this-recursive-arg)))))
      (duplicates-helper (copy-list arg))
      result)))
;; Testing
;; (duplicates '(a a b c d))
;; (duplicates '(a b c d))
;; (duplicates '((a b) (a c) (b c) (d e)))
;; (duplicates '((a b) (a c) (b c) (d e)) :test #'(lambda (x y) (eq (car x) (car y))))

;; returns a list of two lists, where lst1 contains those elements passing
;; the test and lst2 contains the rest of the elements.
(defun split-if (test lst)
  (declare (type (function (*) *) test) (list lst))
  (let* ((pass '())
	 (fail '()))
    (dolist (e lst)
      (if (funcall test e)
	  (push e pass)
	(push e fail)))
    (list pass fail)))

;; Return two values:  the list of input elements passing the test, and
;; the list of those failing the test.
(defun split (test lst)
  (declare (type (function (*) *) test) (list lst))
  (let* ((pass '())
	 (fail '()))
    (dolist (e lst)
      (if (funcall test e)
	  (push e pass)
	(push e fail)))
    (values pass fail)))

;; Argument must be a list.
;; This is more efficient than calling length, which can take a long time
;; to return for long lists.
(defun length-two (list)
  (and (cdr list) (null (cddr list))))
(defun length-one (list)		; also returns true if null
  (null (cdr list)))
;;; Problem: macro expansion seems to take a huge amount of time.
;; (defmacro length-one (l)
;;   `(null (cdr ,l)))
;; (defmacro length-two (l)
;;   (let ((var (gensym)))
;;     `(let ((,var ,l))
;;        (and (cdr ,var) (null (cddr ,var))))))


;; WARNING:  WHEN USING THIS MACRO, PUT A PROGN AROUND IT OR ELSE
;; IT WON'T WORK.
;; Push ITEM onto the list value associated with KEY in ALIST.
;; If KEY doesn't appear in ALIST, then add an entry for it (containing ITEM).
;; Elements of ALIST are (key . item-list) conses, not (key item-list) lists.
;; ALIST is evaluated multiple times, but KEY and ITEM are not.
(defmacro assoc-push (key item alist)
  (let ((key-var (gensym)))
    `(let* ((,key-var ,key)
	    (assoc (assoc ,key-var ,alist)))
       (if assoc
	   (push ,item (cdr assoc))
	 (push (cons ,key-var (list ,item)) ,alist)))))

(defun hashprint (key value)
  (format t "~&~S ~S~%" key value))

(defun htprint (table)
  (maphash #'hashprint table))


(defun string-prefix-p (string prefix)
  (declare (type simple-base-string string) (type simple-base-string prefix))
  (let ((prefixlen (length prefix)))
    (and (>= (length string) prefixlen)
	 (string= prefix string :start2 0 :end2 prefixlen))))

(defun string-suffix-p (string suffix)
  (declare (type simple-base-string string) (type simple-base-string suffix))
  (let ((suffixlen (length suffix)))
    (and (>= (length string) suffixlen)
	 (string= suffix string :start2 (- (length string) suffixlen) :end2 (length string)))))
;; Testing
;; (string-suffix-p "foobar" "bar")
;; (string-suffix-p "bar" "bar")
;; (string-suffix-p "ar" "bar")
;; (string-suffix-p "foofoo" "bar")
;; (string-suffix-p "foo" "bar")

(defmacro with-open-file-maybe ((var file &rest options) &rest body)
  "Like with-open-file, but if FILE is nil, VAR is bound to nil and
no output should be sent to it."
  `(let* ((f ,file)
	  (,var (and f (open f ,@options))))
     (unwind-protect
	 (progn ,@body)
       (if f
	   (close ,var)))))

;;  The largest plan size to consider
(defvar *largest* 30)

;; How many steps to start with
(defvar *start* 1)


;; t if we can do it, nil otherwise
(defun meets-reqs (problem)
  (check-requirements *medic-capabilities* problem nil t))

;; Finds the shortest number of steps
;; for serial and parallel plans
(defun shortest-plan (problem)
  (let ((serial -1)  (parallel -1) (ok (meets-reqs problem)) )
    (progn
      (setq *cnf-solver* 'ubersolve)
      (setq *external-simplifier* nil)
      (setq *find-minimal-plan* t)
      (set-flags "erse")
      (when (equal ok t)
      (setq serial (length (solve-find-problem problem *largest* *start*)))))
    (progn
      (set-flags "erpe")
      (when (equal ok t)
      (setq parallel (length (solve-find-problem problem *largest* *start*)))))
    (cons problem (cons serial (cons parallel () )))))


;; Finds length info for all the problems that don't have it yet.
;; results are written as .length files
(defun find-shortest-plans ()
  (let ( (path nil) (str nil) (name nil) (l nil))
    (dolist (a-prob *problems*)
      (when (not (problem-length (get-problem-copy a-prob)))
	(progn
	  (format t "~%~%Finding length for ~A~%~%" a-prob)
	  (setf l (shortest-plan a-prob))
	  (setf name ( format nil "~A.length" a-prob))
	  (setf path (make-pathname :name name))
	  (setf str (open path :direction :output :if-exists :supersede))
	  (format str "~A  :length ((:serial ~a) (:parallel ~A))~%"(first l) (second l) (third l))
	  (close str))))))


;; Displays the number of problems solvable, lengthified, and total
(defun print-medic-statistics ()
  (let ((number-done 0 ) (number-solveable 0))
    (dolist (a-prob *problems*)
      (when (problem-length (get-problem-copy a-prob))
	(setf number-done ( + 1 number-done)))
      (when (meets-reqs a-prob)
      	(setf number-solveable ( + 1 number-solveable))))

    (format t "~%~%Medic is capable of solving ~A of ~A the problems available." number-solveable (length *problems*))
    (format t "~%~A of ~A problems have length info." number-done (length *problems*))
    ))


;; Verifies that the length information for the problem, if it
;; exists, is correct
(defun verify-length-information ()
    (dolist (a-prob *problems*)
      (when (problem-length (get-problem-copy a-prob))
	(let (  (alleged-parallel-length (nth 1 (assoc :parallel  (problem-length (get-problem-copy a-prob)))))
	      (alleged-serial-length (nth 1 (assoc :serial  (problem-length (get-problem-copy a-prob)))))
	      (actual-length (shortest-plan a-prob)))
	  (progn
	  (when (not (equal (nth 1 actual-length) alleged-serial-length))
	    (error "~%problem ~A serial length is ~a, not ~a as claimed" a-prob (nth 1 actual-length) alleged-serial-length))
	   (when (not (equal (nth 2 actual-length) alleged-parallel-length))
	    (error "~%problem ~A parallel length is ~a, not ~a as claimed" a-prob (nth 2 actual-length) alleged-parallel-length)))))))

;; Lists the problems that don't yet have length info, but are
;; solvable (allegedly) in Medic
(defun list-problems-to-lengthify ()
     (dolist (a-prob *problems*)
      (when (and (not (problem-length (get-problem-copy a-prob))) (meets-reqs a-prob))
      (format t "~%~A" a-prob))))


;; recursively de-nisps a formula/expression/problem
;; that is , converts bindings from (?x - tool) to (tool ?x)
;; and adds colons before and, or, not, etc.
(defun de-nisp (xpr)
  (let ((sub-this '((and . :and) (or . :or) (not . :not)
		     (eq . :eq) (neq . :neq)
		     (when . :when ) (forall . :forall) (exists . :exists))))
  (cond
   ((nisp-expression-p xpr )
      (sublis sub-this (de-nisp-expression xpr)))
   ((listp xpr)
      (sublis sub-this (mapcar #'de-nisp xpr)))
   (t
      (sublis sub-this xpr)))))
;; (de-nisp '(:constants s  - theater d1 d2 - division p1 p2))
;; (de-nisp '(forall (?x - tool ?y) (and (on ?x ?y)  )))




;; Takes a nispy expression (i.e. (?m ?l - location)) and returns a non-nispy
;; expression ( (location ?m) (location ?l))
;; deals with things like (:constants sahara - theater) also,
;; returning (:constants (theater sahara))
(defun de-nisp-expression (xpr)
  (if (not xpr)
      nil
    ;; If xpr is not a list it should be a variable
    (if (not (listp xpr))
      	xpr
    ;; if xpr IS a list, we look through it an item at a time...
    (let ( (var-list nil) (return-list nil) (last-item nil) )
      (dolist (an-item xpr)
	(cond
	 ;; if the last item was a '-, unload var-list
	 ;; and add their bindings to return-list
	 ((equal last-item '-)
	  (let ((temp-list nil) )
	    (dolist (a-var var-list)
	      (if (or (keywordp a-var) (listp a-var))
		  (setf temp-list  (cons a-var temp-list))
		(setf temp-list  (cons (cons  an-item (cons a-var nil)) temp-list))
		))
	    (dolist (b-item temp-list)
	      (setf return-list (cons b-item  return-list))))
	  (setf var-list nil))

	  ( (varp an-item)
	    (setf var-list (cons an-item var-list)))
	  ;; Do nothing with a '-
	  ( (equal '- an-item))
	  ;; if this is a keyword, add it to the var-list
	  ( (keywordp an-item)
	    (setf var-list (cons an-item var-list)))
	  ( t
	    (setf var-list (cons an-item var-list)))
	  )
	 (setf last-item an-item)	; to make sure '- preceeds a symbol
	)

      ;; now we unload the leftovers
        (let ((temp-list nil) )
	    (dolist (a-var var-list)
	      (if (or (keywordp a-var) (listp a-var))
		   (setf temp-list  (cons a-var temp-list))
		(setf temp-list  (cons (cons a-var nil) temp-list))
		))
	    (dolist (b-item temp-list)
	      (setf return-list (cons b-item  return-list))
	      )
	    )

      (reverse return-list)))))



;; (de-nisp-expression '(:constants s  - theater d1 d2 - division p1 p2))
;; (de-nisp-expression '(:forall (?x - tool ?y)))
;; (de-nisp-expression '(?x - tool ?y))
;; (de-nisp-expression '?x)
;; (de-nisp-expression '(?x))
;; (de-nisp-expression '(?x ?y - dog ?z) )
;; (de-nisp-expression '(?y - dog) )
;; (de-nisp-expression '(?x - WORM ?y - CHEESE ?z))
;; (de-nisp-expression '(?x ?y DOG ))

(defun nisp-formula-p (formula)
  "Return t if formula is Nisp style, false otherwise."
  (and formula
       (or (equal formula '- )
	   (and (listp formula)  (nisp-formula-p (car formula)))
	   (and (listp formula)  (nisp-formula-p (cdr formula))))))
;; Tests
;; (nisp-formula-p '(?y - dog) )
;; (nisp-formula-p '(:forall ((?x - tool)) (on ?x table)))
;; (nisp-formula-p '((:or (:eq ?x Table) (:forall (block ?b) (:not (on ?b ?x))))))




(defun nisp-expression-p (formula)
  "Return t if expression is Nisp style, false otherwise."
  (and (listp formula)
       (member '- (cdr formula))))

;; Tests
;; (assert (nisp-expression-p '(?y - dog)))
;; (assert (not (nisp-expression-p '(:forall ((?x - tool)) (on ?x table))))
;; (assert (not (nisp-expression-p '((:or (:eq ?x Table) (:forall (block ?b) (:not (on ?b ?x)))))))
